
class Protobuf(object):
    _protos = dict()
    _names = dict()
    _abbr_names = dict()

    @classmethod
    def populate(cls):
        import re
        from .messages import OpenApiCommonMessages_pb2 as o1
        from .messages import OpenApiMessages_pb2 as o2

        for name in dir(o1) + dir(o2):
            if not name.startswith("Proto"):
                continue

            m = o1 if hasattr(o1, name) else o2
            klass = getattr(m, name)
            cls._protos[klass().payloadType] = klass
            cls._names[klass.__name__] = klass().payloadType
            abbr_name = re.sub(r'^Proto(OA)?(.*)', r'\2', klass.__name__)
            cls._names[abbr_name] = klass().payloadType
        return cls._protos

    @classmethod
    def get(cls, payload, fail=True, **params):
        if not cls._protos:
            cls.populate()

        if payload in cls._protos:
            return cls._protos[payload](**params)

        for d in [cls._names, cls._abbr_names]:
            if payload in d:
                payload = d[payload]
                return cls._protos[payload](**params)
        if fail:  # pragma: nocover
            raise IndexError("Invalid payload: " + str(payload))
        return None  # pragma: nocover

    @classmethod
    def get_type(cls, payload, **params):
        p = cls.get(payload, **params)
        return p.payloadType

    @classmethod
    def extract(cls, message):
        payload = cls.get(message.payloadType)
        payload.ParseFromString(message.payload)
        return payload



# -*- coding: utf-8 -*-

"""Top-level package for spotware_connect."""

__author__ = """Marcus Santos"""
__email__ = 'marcus@marcus-santos.com'
__version__ = '0.1.1'

from .client import Client  # noqa
from .protobuf import Protobuf  # noqa



import sys
from collections import deque
from twisted.protocols.basic import Int32StringReceiver
from twisted.internet import task
from .messages.OpenApiCommonMessages_pb2 \
    import ProtoMessage, ProtoHeartbeatEvent


class Protocol(Int32StringReceiver):

    MAX_LENGTH = sys.maxsize // 2

    _rps_limit = 5
    _send_queue = deque([])
    _send_task_interval = 1
    _send_task = None

    def connectionMade(self):
        super().connectionMade()

        if not self._send_task:
            self._send_task = task.LoopingCall(self._sendStrings)
        self._send_task.start(self._send_task_interval)

    def connectionLost(self, reason):
        super().connectionLost(reason)
        if self._send_task.running:
            self._send_task.stop()

    def heartbeat(self):
        self.send(ProtoHeartbeatEvent(), True)

    def send(self, message, instant=False, msgid=None):
        data = b''

        if isinstance(message, ProtoMessage):
            data = message.SerializeToString()

        if isinstance(message, bytes):
            data = message

        if isinstance(message, ProtoMessage.__base__):
            msg = ProtoMessage(payload=message.SerializeToString(),
                               clientMsgId=msgid,
                               payloadType=message.payloadType)
            data = msg.SerializeToString()

        if instant:
            self.sendString(data)
        else:
            self._send_queue.append(data)

    def _sendStrings(self):
        size = len(self._send_queue)

        if not size:
            return  # pragma: no cover

        for _ in range(min(size, self._rps_limit)):
            self.sendString(self._send_queue.popleft())

    def stringReceived(self, data):
        msg = ProtoMessage()
        msg.ParseFromString(data)

        if msg.payloadType == ProtoHeartbeatEvent().payloadType:
            self.heartbeat()

        self.receive(msg)
        return data

    def receive(self, message):
        pass  # pragma: no cover



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Tests for `spotware_connect` package."""

import unittest
import spotware_connect as sc


class Testspotware_connect(unittest.TestCase):
    """Tests for `spotware_connect` package."""

    _timeout = 5

    def setUp(self):
        """Set up test fixtures, if any."""
        pass

    def tearDown(self):
        """Tear down test fixtures, if any."""
        pass

    def test_connect_api(self):
        """Test basic basic Open Api functions"""
        c = sc.Client()
        self.version = ""

        @c.event
        def connect():
            c.emit("VersionReq")

        @c.message(msgtype="VersionRes")
        def version(version, **kargs):
            self.version = version
            c.stop()

        @c.event
        def disconnect():
            pass

        def receive_heartbeat(protocol):
            from spotware_connect.messages.OpenApiCommonMessages_pb2 \
                import ProtoMessage, ProtoHeartbeatEvent

            payload = ProtoHeartbeatEvent()
            msg = ProtoMessage(payloadType=payload.payloadType,
                               payload=payload.SerializeToString())
            protocol.stringReceived(msg.SerializeToString())

        c.whenConnected().addCallback(receive_heartbeat)
        c.start(timeout=self._timeout)
        self.assertEqual(self.version, "66")



from twisted.internet.protocol import Factory
from twisted.internet.endpoints import clientFromString
from twisted.application.internet import ClientService
from twisted.internet import reactor
from .protocol import Protocol
from .protobuf import Protobuf


class Client(ClientService):
    PROXY_DEMO = "demo.ctraderapi.com:5035"
    PROXY_LIVE = "live.ctraderapi.com:5035"
    EVENT_CONNECT_NAME = "connect"
    EVENT_DISCONNECT_NAME = "disconnect"
    EVENT_MESSAGE_NAME = "message"

    class Protocol(Protocol):
        client = None

        def connectionMade(self):
            super().connectionMade()
            self.client.connect()

        def connectionLost(self, reason):
            super().connectionLost(reason)
            self.client.disconnect()

        def receive(self, message):
            self.client.receive(message)

    class Factory(Factory):
        client = None

        def __init__(self, *args, **kwargs):
            super().__init__()
            self.client = kwargs['client']

        def buildProtocol(self, addr):
            p = super().buildProtocol(addr)
            p.client = self.client
            return p

    def __init__(self, live=False, retryPolicy=None,
                 clock=None, prepareConnection=None):
        host = "ssl:" + (self.PROXY_LIVE if live else self.PROXY_DEMO)
        endpoint = clientFromString(reactor, host)
        factory = Client.Factory.forProtocol(Client.Protocol, client=self)
        super().__init__(endpoint, factory, retryPolicy=retryPolicy,
                         clock=clock, prepareConnection=prepareConnection)

    def start(self, timeout=None):
        self.startService()

        if timeout:
            reactor.callLater(timeout, self.stop)

        reactor.run()

    def stop(self):
        self.stopService()
        if reactor.running:
            reactor.stop()

    def connect(self):
        self.exec_events(self.EVENT_CONNECT_NAME)

    def disconnect(self):
        self.exec_events(self.EVENT_DISCONNECT_NAME)

    def receive(self, message):
        payload = Protobuf.extract(message)
        kargs = dict(msg=message, msgid=message.clientMsgId,
                     msgtype=message.payloadType,
                     payload=payload,
                     **{fv[0].name: fv[1] for fv in payload.ListFields()})

        if "ctidTraderAccountId" in kargs:
            kargs["ctid"] = payload.ctidTraderAccountId

        self.exec_events(self.EVENT_MESSAGE_NAME, **kargs)

    def emit(self, message, msgid=None, **params):
        if type(message) in [str, int]:
            message = Protobuf.get(message, **params)

        def protocol_send(protocol):
            protocol.send(message, msgid=msgid)

        con = self.whenConnected()
        con.addCallback(protocol_send)
        return con

    _events = dict()

    def event(self, name_or_func=None, **filters):
        if not self._events:  # lazy create
            for e in [self.EVENT_CONNECT_NAME,
                      self.EVENT_DISCONNECT_NAME, self.EVENT_MESSAGE_NAME]:
                self._events[e] = []

        if callable(name_or_func):  # callable append
            name = name_or_func.__name__
            self._events[name].append(name_or_func)
            return name_or_func

        def decorate(func):  # decorate with args
            evname = name_or_func

            from functools import wraps

            @wraps(func)
            def func_wrap(*args, **kwargs):
                for k, v in filters.items():
                    if k not in kwargs or kwargs[k] != v:
                        return
                func(*args, **kwargs)

            self._events[evname].append(func_wrap)
            return func

        return decorate

    def message(self, **filters):
        if 'msgtype' in filters and type(filters['msgtype']) in [str, int]:
            filters['msgtype'] = Protobuf.get_type(filters['msgtype'])

        return self.event(self.EVENT_MESSAGE_NAME, **filters)

    def exec_events(self, name, *args, **kwargs):
        if name not in self._events:
            return

        for f in self._events[name]:
            f(*args, **kwargs)
